/*****************************************************************************
            University of West Bohemia, DCS, Pilsen, Czech Republic
                              (c) copyright
                                27.03.2002

                             C-Sim version 5.0

                              file csim_kr.h

            header file of C-Sim kernel based on POSIX threads

*****************************************************************************/

#ifndef CSIM_KR_H
#define CSIM_KR_H 1

#include <pthread.h>
#include "../csim_dt.h"

#define csim_process_kr                                                       \
  pthread_t       thread_id;                                                  \
  pthread_mutex_t mutex;                                                      \
  pthread_cond_t  cond_var;                                                   \
  CSIM_BOOLEAN    thread_run


#define csim_create_process(process)                                          \
do {                                                                          \
  process->thread_run = TRUE;                                                 \
  pthread_mutex_init(&(process->mutex), NULL);                                \
  pthread_cond_init(&(process->cond_var),NULL);                               \
  pthread_mutex_lock(&(csim_sqs_point()->mutex));                             \
  csim_sqs_point()->thread_run = FALSE;                                       \
  if (pthread_create(&(process->thread_id), NULL, process->program,           \
     process) == 0) {                                                         \
    while (csim_sqs_point()->thread_run != TRUE)                              \
      pthread_cond_wait(&(csim_sqs_point()->cond_var),                        \
        &(csim_sqs_point()->mutex));                                          \
    pthread_mutex_unlock(&csim_sqs_point()->mutex);                           \
  }                                                                           \
  else {                                                                      \
    in_step = FALSE;                                                          \
    csim_exception(CSIM_E_CREATE_PROCESS, process, NULL);                     \
  }                                                                           \
} while (0)


#define csim_switch_processes(from, to)                                       \
do {                                                                          \
  CSIM_PROCESS *_p_from, *_p_to;                                              \
  _p_from = (CSIM_PROCESS *) from;                                            \
  _p_to   = (CSIM_PROCESS *) to;                                              \
                                                                              \
  pthread_mutex_lock(&(_p_from->mutex));                                      \
  _p_from->thread_run = FALSE;                                                \
  pthread_mutex_lock(&(_p_to->mutex));                                        \
  _p_to->thread_run = TRUE;                                                   \
  pthread_mutex_unlock(&(_p_to->mutex));                                      \
  pthread_cond_signal(&(_p_to->cond_var));                                    \
                                                                              \
  while (_p_from->thread_run != TRUE)                                         \
    pthread_cond_wait (&(_p_from->cond_var), &(_p_from->mutex));              \
  pthread_mutex_unlock(&(_p_from->mutex));                                    \
} while (0)


#define csim_return_to_step()                                                 \
do {                                                                          \
  pthread_mutex_lock(&(csim_sqs_point()->mutex));                             \
  csim_sqs_point()->thread_run = TRUE;                                        \
  pthread_mutex_unlock(&(csim_sqs_point()->mutex));                           \
  pthread_cond_signal(&(csim_sqs_point()->cond_var));                         \
  pthread_exit(NULL);                                                         \
} while (0)


#define csim_cancel_process(process)                                          \
do {                                                                          \
  pthread_cancel(process->thread_id);                                         \
  pthread_join(process->thread_id, NULL);                                     \
} while (0)

#define csim_init_kr(process)                                                 \
  pthread_mutex_init(&(process->mutex), NULL);                                \
  pthread_cond_init(&(process->cond_var), NULL);                              \
  process->thread_run = TRUE;

#endif
